"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test save module.
"""
from pathlib import Path
import unittest
from shutil import rmtree

import numpy as np

from climada import CONFIG
from climada.util.api_client import Client, DataTypeShortInfo, Download

DATA_DIR = CONFIG.test_data.dir()


class TestClient(unittest.TestCase):
    """Test data api client methods."""

    def test_data_type(self):
        """"""
        lpdt = Client().get_data_type_info("tropical_cyclone")
        self.assertEqual(lpdt.data_type, 'tropical_cyclone')
        self.assertEqual(lpdt.data_type_group, 'hazard')
        self.assertTrue('res_arcsec' in [p['property'] for p in lpdt.properties if p['mandatory']])
        self.assertTrue('ref_year' in [p['property'] for p in lpdt.properties if not p['mandatory']])

    def test_data_types(self):
        """"""
        exdts = Client().list_data_type_infos("exposures")
        self.assertTrue('litpop' in [exdt.data_type for exdt in exdts])

    def test_datasets(self):
        """"""
        datasets = Client().list_dataset_infos(status=None, name='FAOSTAT_data_producer_prices')
        self.assertEqual(len(datasets), 1)

    def test_dataset(self):
        """"""
        client = Client()

        dataset = client.get_dataset_info(name='FAOSTAT_data_producer_prices', status='test_dataset')
        self.assertEqual(dataset.version, 'v1')
        self.assertEqual(len(dataset.files), 1)
        self.assertEqual(dataset.files[0].file_size, 26481)
        self.assertEqual(dataset.data_type, DataTypeShortInfo('crop_production', 'exposures'))

        dataset2 = client.get_dataset_info_by_uuid(dataset.uuid)
        self.assertEqual(dataset, dataset2)

    def test_download_file(self):
        """"""
        client = Client()
        client.MAX_WAITING_PERIOD = 0.1
        dataset = client.get_dataset_info(name='FAOSTAT_data_producer_prices', status='test_dataset')

        # test failure
        def fail(x, y):
            raise Download.Failed("on purpose")
        self.assertRaises(Download.Failed,
            client._download_file, DATA_DIR, dataset.files[0], check=fail)
        self.assertFalse(DATA_DIR.joinpath(dataset.files[0].file_name).is_file())

        # test success
        download = client._download_file(DATA_DIR, dataset.files[0])
        self.assertEqual(download, DATA_DIR / dataset.files[0].file_name)
        self.assertTrue(download.is_file())
        self.assertEqual(download.stat().st_size, dataset.files[0].file_size)
        download.unlink()
        self.assertFalse(download.is_file())

    def test_download_dataset(self):
        """"""
        client = Client()
        client.MAX_WAITING_PERIOD = 0.1

        dataset = client.get_dataset_info(name='test_write_raster', status='test_dataset')
        download_dir, downloads = client.download_dataset(dataset, target_dir=DATA_DIR)
        self.assertEqual(download_dir.name, dataset.version)
        self.assertEqual(download_dir.parent.name, dataset.name)
        self.assertEqual(download_dir.parent.parent.name, dataset.data_type.data_type)
        self.assertEqual(len(downloads), 2)
        for download, dsfile in zip(downloads, dataset.files):
            self.assertTrue(download.is_file())
            self.assertEqual(download.stat().st_size, dsfile.file_size)
            self.assertEqual(download.name, dsfile.file_name)
            self.assertEqual(download.parent, download_dir)
            download.unlink()
        rm_empty_dir(download.parent.parent.parent)

    def test_get_exposures(self):
        client = Client()
        exposures = client.get_exposures(exposures_type='litpop',
                                         properties={'country_iso3alpha': 'AUT',
                                                     'fin_mode': 'pop', 'exponents': '(0,1)'},
                                         dump_dir=DATA_DIR)
        self.assertEqual(len(exposures.gdf), 5782)
        self.assertEqual(np.unique(exposures.gdf.region_id), 40)
        self.assertTrue('[0, 1]' in exposures.tag.description)
        self.assertTrue('pop' in exposures.tag.description)
        exposures

    def test_get_exposures_fails(self):
        client = Client()
        with self.assertRaises(ValueError) as cm:
            client.get_exposures(exposures_type='river_flood', 
                                 properties={'country_iso3alpha': 'AUT',
                                             'fin_mode': 'pop', 'exponents': '(0,1)'},
                                 dump_dir=DATA_DIR)
        self.assertIn('Valid exposures types are a subset of CLIMADA exposures types. Currently',
                      str(cm.exception))

        with self.assertRaises(Client.AmbiguousResult) as cm:
            client.get_exposures(exposures_type='litpop', 
                                 properties={'country_iso3alpha': 'AUT'},
                                 dump_dir=DATA_DIR)
        self.assertIn('there are several datasets meeting the requirements',
                      str(cm.exception))

    def test_get_hazard(self):
        client = Client()
        hazard = client.get_hazard(hazard_type='river_flood',
                                   properties={'country_name': 'Austria',
                                               'year_range': '2010_2030', 'climate_scenario': 'rcp26'},
                                   dump_dir=DATA_DIR)
        self.assertEqual(np.shape(hazard.intensity), (480, 5784))
        self.assertEqual(np.unique(hazard.centroids.region_id), 40)
        self.assertEqual(np.unique(hazard.date).size, 20)
        self.assertEqual(hazard.tag.haz_type, 'RF')

    def test_get_hazard_fails(self):
        client = Client()
        with self.assertRaises(ValueError) as cm:
            client.get_hazard(hazard_type='litpop', 
                              properties={'country_name': 'Austria',
                                          'year_range': '2010_2030', 'climate_scenario': 'rcp26'},
                              dump_dir=DATA_DIR)
        self.assertIn('Valid hazard types are a subset of CLIMADA hazard types. Currently',
                      str(cm.exception))

        with self.assertRaises(Client.AmbiguousResult) as cm:
            client.get_hazard(hazard_type='river_flood',
                              properties={'country_name': ['Switzerland', 'Austria'],
                                          'year_range': '2010_2030', 'climate_scenario': ['rcp26', 'rcp85']},
                              dump_dir=DATA_DIR)
        self.assertIn('there are several datasets meeting the requirements:', str(cm.exception))

    def test_get_litpop_default(self):
        client = Client()
        litpop = client.get_litpop_default(country='LUX', dump_dir=DATA_DIR)
        self.assertEqual(len(litpop.gdf), 188)
        self.assertEqual(np.unique(litpop.gdf.region_id), 442)
        self.assertTrue('[1, 1]' in litpop.tag.description)
        self.assertTrue('pc' in litpop.tag.description)

    def test_multi_filter(self):
        client = Client()
        testds = client.list_dataset_infos(data_type='storm_europe')

        # assert no systemic loss in filtering
        still = client._filter_datasets(testds, dict())
        for o, r in zip(testds, still):
            self.assertEqual(o, r)

        # assert filter is effective
        p = 'country_name'
        a, b = 'Germany', 'Netherlands'
        less = client._filter_datasets(testds, {p:[a, b]})
        self.assertLess(len(less), len(testds))
        only = client._filter_datasets(testds, {p:[a]})
        self.assertLess(len(only), len(less))
        self.assertLess(0, len(only))

    def test_multiplicity_split(self):
        properties = {
            'country_name': ['x', 'y', 'z'],
            'b': '1'
        }
        # assert split matches expectations
        straight, multi = Client._divide_straight_from_multi(properties)
        self.assertEqual(straight, {'b': '1'})
        self.assertEqual(multi, {'country_name': ['x', 'y', 'z']})


def rm_empty_dir(folder):
    for subfolder in folder.iterdir():
        if subfolder.is_dir():
            rm_empty_dir(subfolder)
    try:
        folder.rmdir()
    except:
        pass


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestClient)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test ImpactFuncSet class.
"""
import unittest
import numpy as np

from climada import CONFIG
from climada.entity.impact_funcs.impact_func_set import ImpactFuncSet, ImpactFunc
from climada.util.constants import ENT_TEMPLATE_XLS, ENT_DEMO_TODAY

ENT_TEST_MAT = CONFIG.exposures.test_data.dir().joinpath('demo_today.mat')

class TestConstructor(unittest.TestCase):
    """Test impact function attributes."""
    def test_attributes_all(self):
        """All attributes are defined"""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.haz_type = 'TC'
        vulner_1.id = '2'
        self.assertTrue(hasattr(imp_fun, 'tag'))
        self.assertTrue(hasattr(imp_fun, '_data'))
        self.assertTrue(hasattr(vulner_1, 'haz_type'))
        self.assertTrue(hasattr(vulner_1, 'name'))
        self.assertTrue(hasattr(vulner_1, 'id'))
        self.assertTrue(hasattr(vulner_1, 'intensity_unit'))
        self.assertTrue(hasattr(vulner_1, 'mdd'))
        self.assertTrue(hasattr(vulner_1, 'paa'))

class TestContainer(unittest.TestCase):
    """Test ImpactFuncSet as container."""
    def test_add_wrong_error(self):
        """Test error is raised when wrong ImpactFunc provided."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        with self.assertLogs('climada.entity.impact_funcs.impact_func_set',
                             level='WARNING') as cm:
            imp_fun.append(vulner_1)
        self.assertIn("Input ImpactFunc's hazard type not set.", cm.output[0])

        vulner_1.haz_type = 'TC'
        with self.assertLogs('climada.entity.impact_funcs.impact_func_set',
                             level='WARNING') as cm:
            imp_fun.append(vulner_1)
        self.assertIn("Input ImpactFunc's id not set.", cm.output[0])

        with self.assertRaises(ValueError) as cm:
            imp_fun.append(45)
        self.assertIn("Input value is not of type ImpactFunc.", str(cm.exception))

    def test_remove_func_pass(self):
        """Test remove_func removes ImpactFunc of ImpactFuncSet correcty."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)
        imp_fun.remove_func()
        self.assertEqual(0, len(imp_fun._data))

    def test_remove_wrong_error(self):
        """Test error is raised when invalid inputs."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)
        with self.assertLogs('climada.entity.impact_funcs.impact_func_set', level='WARNING') as cm:
            imp_fun.remove_func('FL')
        self.assertIn('No ImpactFunc with hazard FL.', cm.output[0])
        with self.assertLogs('climada.entity.impact_funcs.impact_func_set', level='WARNING') as cm:
            imp_fun.remove_func(fun_id=3)
        self.assertIn('No ImpactFunc with id 3.', cm.output[0])

    def test_get_hazards_pass(self):
        """Test get_hazard_types function."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)
        self.assertEqual(1, len(imp_fun.get_hazard_types()))
        self.assertEqual(['TC'], imp_fun.get_hazard_types())

        vulner_2 = ImpactFunc()
        vulner_2.id = 1
        vulner_2.haz_type = 'TC'
        imp_fun.append(vulner_2)
        self.assertEqual(1, len(imp_fun.get_hazard_types()))
        self.assertEqual(['TC'], imp_fun.get_hazard_types())

        vulner_3 = ImpactFunc()
        vulner_3.id = 1
        vulner_3.haz_type = 'FL'
        imp_fun.append(vulner_3)
        self.assertEqual(2, len(imp_fun.get_hazard_types()))
        self.assertIn('TC', imp_fun.get_hazard_types())
        self.assertIn('FL', imp_fun.get_hazard_types())

    def test_get_ids_pass(self):
        """Test normal functionality of get_ids method."""
        imp_fun = ImpactFuncSet()
        self.assertEqual({}, imp_fun.get_ids())

        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)
        self.assertEqual(1, len(imp_fun.get_ids()))
        self.assertIn('TC', imp_fun.get_ids())
        self.assertEqual(1, len(imp_fun.get_ids('TC')))
        self.assertEqual([1], imp_fun.get_ids('TC'))

        vulner_2 = ImpactFunc()
        vulner_2.id = 3
        vulner_2.haz_type = 'TC'
        imp_fun.append(vulner_2)
        self.assertEqual(1, len(imp_fun.get_ids()))
        self.assertIn('TC', imp_fun.get_ids())
        self.assertEqual(2, len(imp_fun.get_ids('TC')))
        self.assertEqual([1, 3], imp_fun.get_ids('TC'))

        vulner_3 = ImpactFunc()
        vulner_3.id = 3
        vulner_3.haz_type = 'FL'
        imp_fun.append(vulner_3)
        self.assertEqual(2, len(imp_fun.get_ids()))
        self.assertIn('TC', imp_fun.get_ids())
        self.assertIn('FL', imp_fun.get_ids())
        self.assertEqual(2, len(imp_fun.get_ids('TC')))
        self.assertEqual([1, 3], imp_fun.get_ids('TC'))
        self.assertEqual(1, len(imp_fun.get_ids('FL')))
        self.assertEqual([3], imp_fun.get_ids('FL'))

    def test_get_ids_wrong_zero(self):
        """Test get_ids method with wrong inputs."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.haz_type = 'WS'
        vulner_1.id = 56
        imp_fun.append(vulner_1)
        self.assertEqual([], imp_fun.get_ids('TC'))

    def test_get_func_pass(self):
        """Test normal functionality of get_func method."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.haz_type = 'WS'
        vulner_1.id = 56
        imp_fun.append(vulner_1)
        self.assertEqual(1, len(imp_fun.get_func('WS')))
        self.assertEqual(1, len(imp_fun.get_func(fun_id=56)))
        self.assertIs(vulner_1, imp_fun.get_func('WS', 56))

        vulner_2 = ImpactFunc()
        vulner_2.haz_type = 'WS'
        vulner_2.id = 6
        imp_fun.append(vulner_2)
        self.assertEqual(2, len(imp_fun.get_func('WS')))
        self.assertEqual(1, len(imp_fun.get_func(fun_id=6)))
        self.assertIs(vulner_2, imp_fun.get_func('WS', 6))

        vulner_3 = ImpactFunc()
        vulner_3.haz_type = 'TC'
        vulner_3.id = 6
        imp_fun.append(vulner_3)
        self.assertEqual(2, len(imp_fun.get_func(fun_id=6)))
        self.assertEqual(1, len(imp_fun.get_func(fun_id=56)))
        self.assertEqual(2, len(imp_fun.get_func('WS')))
        self.assertEqual(1, len(imp_fun.get_func('TC')))
        self.assertIs(vulner_3, imp_fun.get_func('TC', 6))

        self.assertEqual(2, len(imp_fun.get_func().keys()))
        self.assertEqual(1, len(imp_fun.get_func()['TC'].keys()))
        self.assertEqual(2, len(imp_fun.get_func()['WS'].keys()))

    def test_get_func_wrong_error(self):
        """Test get_func method with wrong inputs."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.haz_type = 'WS'
        vulner_1.id = 56
        imp_fun.append(vulner_1)
        self.assertEqual([], imp_fun.get_func('TC'))

    def test_size_pass(self):
        """Test size function."""
        imp_fun = ImpactFuncSet()
        self.assertEqual(0, imp_fun.size())

        vulner_1 = ImpactFunc()
        vulner_1.haz_type = 'WS'
        vulner_1.id = 56
        imp_fun.append(vulner_1)
        self.assertEqual(1, imp_fun.size())
        self.assertEqual(1, imp_fun.size('WS', 56))
        self.assertEqual(1, imp_fun.size('WS'))
        self.assertEqual(1, imp_fun.size(fun_id=56))
        imp_fun.append(vulner_1)
        self.assertEqual(1, imp_fun.size())
        self.assertEqual(1, imp_fun.size('WS', 56))
        self.assertEqual(1, imp_fun.size('WS'))
        self.assertEqual(1, imp_fun.size(fun_id=56))

        vulner_2 = ImpactFunc()
        vulner_2.haz_type = 'WS'
        vulner_2.id = 5
        imp_fun.append(vulner_2)
        self.assertEqual(2, imp_fun.size())
        self.assertEqual(1, imp_fun.size('WS', 56))
        self.assertEqual(2, imp_fun.size('WS'))
        self.assertEqual(1, imp_fun.size(fun_id=56))
        self.assertEqual(1, imp_fun.size(fun_id=5))

        vulner_3 = ImpactFunc()
        vulner_3.haz_type = 'TC'
        vulner_3.id = 5
        imp_fun.append(vulner_3)
        self.assertEqual(3, imp_fun.size())
        self.assertEqual(1, imp_fun.size('TC', 5))
        self.assertEqual(2, imp_fun.size('WS'))
        self.assertEqual(1, imp_fun.size('TC'))
        self.assertEqual(1, imp_fun.size(fun_id=56))
        self.assertEqual(2, imp_fun.size(fun_id=5))

    def test_size_wrong_zero(self):
        """Test size method with wrong inputs."""
        imp_fun = ImpactFuncSet()
        self.assertEqual(0, imp_fun.size('TC'))
        self.assertEqual(0, imp_fun.size('TC', 3))
        self.assertEqual(0, imp_fun.size(fun_id=3))

    def test_append_pass(self):
        """Test append adds ImpactFunc to ImpactFuncSet correctly."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)
        self.assertEqual(1, len(imp_fun._data))
        self.assertIn('TC', imp_fun._data.keys())
        self.assertEqual(1, len(imp_fun._data['TC']))
        self.assertIn(1, imp_fun._data['TC'].keys())

        vulner_2 = ImpactFunc()
        vulner_2.id = 3
        vulner_2.haz_type = 'TC'
        imp_fun.append(vulner_2)
        self.assertEqual(1, len(imp_fun._data))
        self.assertIn('TC', imp_fun._data.keys())
        self.assertEqual(2, len(imp_fun._data['TC']))
        self.assertIn(1, imp_fun._data['TC'].keys())
        self.assertIn(3, imp_fun._data['TC'].keys())

        vulner_3 = ImpactFunc()
        vulner_3.id = 3
        vulner_3.haz_type = 'FL'
        imp_fun.append(vulner_3)
        self.assertEqual(2, len(imp_fun._data))
        self.assertIn('TC', imp_fun._data.keys())
        self.assertIn('FL', imp_fun._data.keys())
        self.assertEqual(2, len(imp_fun._data['TC']))
        self.assertEqual(1, len(imp_fun._data['FL']))
        self.assertIn(1, imp_fun._data['TC'].keys())
        self.assertIn(3, imp_fun._data['TC'].keys())
        self.assertIn(3, imp_fun._data['FL'].keys())

    def test_remove_add_pass(self):
        """Test ImpactFunc can be added after removing."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)
        imp_fun.remove_func()
        self.assertEqual(0, len(imp_fun.get_hazard_types()))
        self.assertEqual(0, len(imp_fun.get_ids()))

        imp_fun.append(vulner_1)
        self.assertEqual(1, len(imp_fun.get_hazard_types()))
        self.assertEqual('TC', imp_fun.get_hazard_types()[0])
        self.assertEqual(1, len(imp_fun.get_ids()))
        self.assertEqual([1], imp_fun.get_ids('TC'))

class TestChecker(unittest.TestCase):
    """Test loading funcions from the ImpactFuncSet class"""
    def test_check_wrongPAA_fail(self):
        """Wrong PAA definition"""
        imp_fun = ImpactFuncSet()
        vulner = ImpactFunc()
        vulner.id = 1
        vulner.haz_type = 'TC'
        vulner.intensity = np.array([1, 2, 3])
        vulner.mdd = np.array([1, 2, 3])
        vulner.paa = np.array([1, 2])
        imp_fun.append(vulner)

        with self.assertRaises(ValueError) as cm:
            imp_fun.check()
        self.assertIn('Invalid ImpactFunc.paa size: 3 != 2.', str(cm.exception))

    def test_check_wrongMDD_fail(self):
        """Wrong MDD definition"""
        imp_fun = ImpactFuncSet()
        vulner = ImpactFunc()
        vulner.id = 1
        vulner.haz_type = 'TC'
        vulner.intensity = np.array([1, 2, 3])
        vulner.mdd = np.array([1, 2])
        vulner.paa = np.array([1, 2, 3])
        imp_fun.append(vulner)

        with self.assertRaises(ValueError) as cm:
            imp_fun.check()
        self.assertIn('Invalid ImpactFunc.mdd size: 3 != 2.', str(cm.exception))

class TestExtend(unittest.TestCase):
    """Check extend function"""
    def test_extend_to_empty_same(self):
        """Extend ImpactFuncSet to empty one."""
        imp_fun = ImpactFuncSet()
        imp_fun_add = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun_add.append(vulner_1)

        vulner_2 = ImpactFunc()
        vulner_2.id = 3
        vulner_2.haz_type = 'TC'
        imp_fun_add.append(vulner_2)

        vulner_3 = ImpactFunc()
        vulner_3.id = 3
        vulner_3.haz_type = 'FL'
        imp_fun_add.append(vulner_3)

        imp_fun_add.tag.file_name = 'file1.txt'

        imp_fun.extend(imp_fun_add)
        imp_fun.check()

        self.assertEqual(imp_fun.size(), 3)
        self.assertEqual(imp_fun.size('TC'), 2)
        self.assertEqual(imp_fun.size('FL'), 1)
        self.assertEqual(imp_fun.tag.file_name, imp_fun_add.tag.file_name)
        self.assertEqual(imp_fun.tag.description, imp_fun_add.tag.description)

    def test_extend_equal_same(self):
        """Extend the same ImpactFuncSet. The inital ImpactFuncSet is obtained."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)

        imp_fun_add = ImpactFuncSet()
        imp_fun_add.append(vulner_1)

        imp_fun.extend(imp_fun_add)
        imp_fun.check()

        self.assertEqual(imp_fun.size(), 1)
        self.assertEqual(imp_fun.size('TC'), 1)

    def test_extend_different_extend(self):
        """Extend ImpactFuncSet with same and new values. The vulnerabilities
        with repeated id are overwritten."""
        imp_fun = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun.append(vulner_1)

        vulner_2 = ImpactFunc()
        vulner_2.id = 3
        vulner_2.haz_type = 'TC'
        imp_fun.append(vulner_2)

        vulner_3 = ImpactFunc()
        vulner_3.id = 3
        vulner_3.haz_type = 'FL'
        imp_fun.append(vulner_3)

        imp_fun_add = ImpactFuncSet()
        vulner_1 = ImpactFunc()
        vulner_1.id = 1
        vulner_1.haz_type = 'TC'
        imp_fun_add.append(vulner_1)

        vulner_2 = ImpactFunc()
        vulner_2.id = 1
        vulner_2.haz_type = 'WS'
        imp_fun_add.append(vulner_2)

        vulner_3 = ImpactFunc()
        vulner_3.id = 3
        vulner_3.haz_type = 'FL'
        imp_fun_add.append(vulner_3)

        imp_fun.extend(imp_fun_add)
        imp_fun.check()

        self.assertEqual(imp_fun.size(), 4)
        self.assertEqual(imp_fun.size('TC'), 2)
        self.assertEqual(imp_fun.size('FL'), 1)
        self.assertEqual(imp_fun.size('WS'), 1)

class TestReaderMat(unittest.TestCase):
    """Test reader functionality of the imp_funcsFuncsExcel class"""

    def test_demo_file_pass(self):
        """Read demo excel file"""
        # Read demo mat file
        description = 'One single file.'
        imp_funcs = ImpactFuncSet.from_mat(ENT_TEST_MAT, description)

        # Check results
        n_funcs = 2
        hazard = 'TC'
        first_id = 1
        second_id = 3

        self.assertEqual(len(imp_funcs._data), 1)
        self.assertEqual(len(imp_funcs._data[hazard]), n_funcs)

        # first function
        self.assertEqual(imp_funcs._data[hazard][first_id].id, 1)
        self.assertEqual(imp_funcs._data[hazard][first_id].name,
                         'Tropical cyclone default')
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity_unit,
                         'm/s')

        self.assertEqual(imp_funcs._data[hazard][first_id].intensity.shape,
                         (9,))
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[0], 0)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[1], 20)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[2], 30)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[3], 40)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[4], 50)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[5], 60)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[6], 70)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[7], 80)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[8], 100)

        self.assertEqual(imp_funcs._data[hazard][first_id].mdd.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][first_id].mdd[0], 0)
        self.assertEqual(imp_funcs._data[hazard][first_id].mdd[8], 0.41079600)

        self.assertEqual(imp_funcs._data[hazard][first_id].paa.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][first_id].paa[0], 0)
        self.assertEqual(imp_funcs._data[hazard][first_id].paa[8], 1)

        # second function
        self.assertEqual(imp_funcs._data[hazard][second_id].id, 3)
        self.assertEqual(imp_funcs._data[hazard][second_id].name,
                         'TC Building code')
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity_unit,
                         'm/s')

        self.assertEqual(imp_funcs._data[hazard][second_id].intensity.shape,
                         (9,))
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[0], 0)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[1], 20)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[2], 30)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[3], 40)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[4], 50)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[5], 60)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[6], 70)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[7], 80)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[8], 100)

        self.assertEqual(imp_funcs._data[hazard][second_id].mdd.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][second_id].mdd[0], 0)
        self.assertEqual(imp_funcs._data[hazard][second_id].mdd[8], 0.4)

        self.assertEqual(imp_funcs._data[hazard][second_id].paa.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][second_id].paa[0], 0)
        self.assertEqual(imp_funcs._data[hazard][second_id].paa[8], 1)

        # general information
        self.assertEqual(imp_funcs.tag.file_name, str(ENT_TEST_MAT))
        self.assertEqual(imp_funcs.tag.description, description)

class TestReaderExcel(unittest.TestCase):
    """Test reader functionality of the imp_funcsFuncsExcel class"""

    def test_demo_file_pass(self):
        """Read demo excel file"""
        # Read demo excel file

        description = 'One single file.'
        imp_funcs = ImpactFuncSet.from_excel(ENT_DEMO_TODAY, description)

        # Check results
        n_funcs = 2
        hazard = 'TC'
        first_id = 1
        second_id = 3

        self.assertEqual(len(imp_funcs._data), 1)
        self.assertEqual(len(imp_funcs._data[hazard]), n_funcs)

        # first function
        self.assertEqual(imp_funcs._data[hazard][first_id].id, 1)
        self.assertEqual(imp_funcs._data[hazard][first_id].name,
                         'Tropical cyclone default')
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity_unit,
                         'm/s')

        self.assertEqual(imp_funcs._data[hazard][first_id].intensity.shape,
                         (9,))
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[0], 0)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[1], 20)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[2], 30)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[3], 40)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[4], 50)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[5], 60)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[6], 70)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[7], 80)
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity[8], 100)

        self.assertEqual(imp_funcs._data[hazard][first_id].mdd.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][first_id].mdd[0], 0)
        self.assertEqual(imp_funcs._data[hazard][first_id].mdd[8], 0.41079600)

        self.assertEqual(imp_funcs._data[hazard][first_id].paa.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][first_id].paa[0], 0)
        self.assertEqual(imp_funcs._data[hazard][first_id].paa[8], 1)

        # second function
        self.assertEqual(imp_funcs._data[hazard][second_id].id, 3)
        self.assertEqual(imp_funcs._data[hazard][second_id].name,
                         'TC Building code')
        self.assertEqual(imp_funcs._data[hazard][first_id].intensity_unit,
                         'm/s')

        self.assertEqual(imp_funcs._data[hazard][second_id].intensity.shape,
                         (9,))
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[0], 0)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[1], 20)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[2], 30)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[3], 40)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[4], 50)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[5], 60)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[6], 70)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[7], 80)
        self.assertEqual(imp_funcs._data[hazard][second_id].intensity[8], 100)

        self.assertEqual(imp_funcs._data[hazard][second_id].mdd.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][second_id].mdd[0], 0)
        self.assertEqual(imp_funcs._data[hazard][second_id].mdd[8], 0.4)

        self.assertEqual(imp_funcs._data[hazard][second_id].paa.shape, (9,))
        self.assertEqual(imp_funcs._data[hazard][second_id].paa[0], 0)
        self.assertEqual(imp_funcs._data[hazard][second_id].paa[8], 1)

        # general information
        self.assertEqual(imp_funcs.tag.file_name, str(ENT_DEMO_TODAY))
        self.assertEqual(imp_funcs.tag.description, description)

    def test_template_file_pass(self):
        """Read template excel file"""
        imp_funcs = ImpactFuncSet.from_excel(ENT_TEMPLATE_XLS)
        # Check some results
        self.assertEqual(len(imp_funcs._data), 10)
        self.assertEqual(len(imp_funcs._data['TC'][3].paa), 9)
        self.assertEqual(len(imp_funcs._data['EQ'][1].intensity), 14)
        self.assertEqual(len(imp_funcs._data['HS'][1].mdd), 16)

class TestWriter(unittest.TestCase):
    """Test reader functionality of the imp_funcsFuncsExcel class"""

    def test_write_read_pass(self):
        """Write + read excel file"""

        imp_funcs = ImpactFuncSet()
        imp_funcs.tag.file_name = 'No file name'
        imp_funcs.tag.description = 'test writer'

        imp1 = ImpactFunc()
        imp1.id = 1
        imp1.name = 'code 1'
        imp1.intensity_unit = 'm/s'
        imp1.haz_type = 'TC'
        imp1.intensity = np.arange(100)
        imp1.mdd = np.arange(100) * 0.5
        imp1.paa = np.ones(100)
        imp_funcs.append(imp1)

        imp2 = ImpactFunc()
        imp2.id = 2
        imp2.name = 'code 2'
        imp2.intensity_unit = 'm/s'
        imp2.haz_type = 'TC'
        imp2.intensity = np.arange(102)
        imp2.mdd = np.arange(102) * 0.25
        imp2.paa = np.ones(102)
        imp_funcs.append(imp2)

        imp3 = ImpactFunc()
        imp3.id = 1
        imp3.name = 'code 1'
        imp3.intensity_unit = 'm'
        imp3.haz_type = 'FL'
        imp3.intensity = np.arange(86)
        imp3.mdd = np.arange(86) * 0.15
        imp3.paa = np.ones(86)
        imp_funcs.append(imp3)

        imp4 = ImpactFunc()
        imp4.id = 15
        imp4.name = 'code 15'
        imp4.intensity_unit = 'K'
        imp4.haz_type = 'DR'
        imp4.intensity = np.arange(5)
        imp4.mdd = np.arange(5)
        imp4.paa = np.ones(5)
        imp_funcs.append(imp4)

        file_name = CONFIG.impact_funcs.test_data.dir().joinpath('test_write.xlsx')
        imp_funcs.write_excel(file_name)

        imp_res = ImpactFuncSet.from_excel(file_name)

        self.assertEqual(imp_res.tag.file_name, str(file_name))
        self.assertEqual(imp_res.tag.description, '')

        # first function
        for fun_haz, fun_dict in imp_res.get_func().items():
            for fun_id, fun in fun_dict.items():
                if fun_haz == 'TC' and fun_id == 1:
                    ref_fun = imp1
                elif fun_haz == 'TC' and fun_id == 2:
                    ref_fun = imp2
                elif fun_haz == 'FL' and fun_id == 1:
                    ref_fun = imp3
                elif fun_haz == 'DR' and fun_id == 15:
                    ref_fun = imp4
                else:
                    self.assertEqual(1, 0)

                self.assertEqual(ref_fun.haz_type, fun.haz_type)
                self.assertEqual(ref_fun.id, fun.id)
                self.assertEqual(ref_fun.name, fun.name)
                self.assertEqual(ref_fun.intensity_unit, fun.intensity_unit)
                self.assertTrue(np.allclose(ref_fun.intensity, fun.intensity))
                self.assertTrue(np.allclose(ref_fun.mdd, fun.mdd))
                self.assertTrue(np.allclose(ref_fun.paa, fun.paa))

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestContainer)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestChecker))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestExtend))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestReaderExcel))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestReaderMat))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestWriter))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestConstructor))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Hazard base class.
"""

import unittest
import numpy as np
from scipy import sparse

from climada import CONFIG
from climada.hazard.base import Hazard
from climada.hazard.centroids import Centroids
from climada.util.constants import HAZ_DEMO_FL

DATA_DIR = CONFIG.test_data.dir()

class TestCentroids(unittest.TestCase):
    """Test centroids functionalities"""

    def test_read_write_raster_pass(self):
        """Test write_raster: Hazard from raster data"""
        haz_fl = Hazard.from_raster([HAZ_DEMO_FL])
        haz_fl.tag.haz_type = 'FL'
        haz_fl.check()

        self.assertEqual(haz_fl.intensity.shape, (1, 1032226))
        self.assertEqual(haz_fl.intensity.min(), -9999.0)
        self.assertAlmostEqual(haz_fl.intensity.max(), 4.662774085998535)

        haz_fl.write_raster(DATA_DIR.joinpath('test_write_hazard.tif'))

        haz_read = Hazard.from_raster([DATA_DIR.joinpath('test_write_hazard.tif')])
        haz_fl.tag.haz_type = 'FL'
        self.assertTrue(np.allclose(haz_fl.intensity.toarray(), haz_read.intensity.toarray()))
        self.assertEqual(np.unique(np.array(haz_fl.fraction.toarray())).size, 2)

    def test_read_raster_pool_pass(self):
        """Test from_raster constructor with pool"""
        from pathos.pools import ProcessPool as Pool
        pool = Pool()
        haz_fl = Hazard.from_raster([HAZ_DEMO_FL], haz_type='FL', pool=pool)
        haz_fl.check()

        self.assertEqual(haz_fl.intensity.shape, (1, 1032226))
        self.assertEqual(haz_fl.intensity.min(), -9999.0)
        self.assertAlmostEqual(haz_fl.intensity.max(), 4.662774085998535)
        pool.close()
        pool.join()

    def test_read_write_vector_pass(self):
        """Test write_raster: Hazard from vector data"""
        haz_fl = Hazard('FL')
        haz_fl.event_id = np.array([1])
        haz_fl.date = np.array([1])
        haz_fl.frequency = np.array([1])
        haz_fl.orig = np.array([1])
        haz_fl.event_name = ['1']
        haz_fl.intensity = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]))
        haz_fl.fraction = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]) / 2)
        haz_fl.centroids = Centroids.from_lat_lon(np.array([1, 2, 3]), np.array([1, 2, 3]))
        haz_fl.check()

        haz_fl.write_raster(DATA_DIR.joinpath('test_write_hazard.tif'))

        haz_read = Hazard.from_raster([DATA_DIR.joinpath('test_write_hazard.tif')], haz_type='FL')
        self.assertEqual(haz_read.intensity.shape, (1, 9))
        self.assertTrue(np.allclose(np.unique(np.array(haz_read.intensity.toarray())),
                                    np.array([0.0, 0.1, 0.2, 0.5])))

    def test_write_fraction_pass(self):
        """Test write_raster with fraction"""
        haz_fl = Hazard('FL')
        haz_fl.event_id = np.array([1])
        haz_fl.date = np.array([1])
        haz_fl.frequency = np.array([1])
        haz_fl.orig = np.array([1])
        haz_fl.event_name = ['1']
        haz_fl.intensity = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]))
        haz_fl.fraction = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]) / 2)
        haz_fl.centroids = Centroids.from_lat_lon(np.array([1, 2, 3]), np.array([1, 2, 3]))
        haz_fl.check()

        haz_fl.write_raster(DATA_DIR.joinpath('test_write_hazard.tif'), intensity=False)

        haz_read = Hazard.from_raster([DATA_DIR.joinpath('test_write_hazard.tif')],
                                      files_fraction=[DATA_DIR.joinpath('test_write_hazard.tif')],
                                      haz_type='FL')
        self.assertEqual(haz_read.intensity.shape, (1, 9))
        self.assertEqual(haz_read.fraction.shape, (1, 9))
        self.assertTrue(np.allclose(np.unique(np.array(haz_read.fraction.toarray())),
                                    np.array([0.0, 0.05, 0.1, 0.25])))
        self.assertTrue(np.allclose(np.unique(np.array(haz_read.intensity.toarray())),
                                    np.array([0.0, 0.05, 0.1, 0.25])))

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestCentroids)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test variable check util functions.
"""

import unittest
import numpy as np
import scipy.sparse as sparse

import climada.util.checker as u_check

class DummyClass(object):

    vars_oblig = {'id', 'array', 'sparse_arr'}
    vars_opt = {'list', 'array_opt'}

    def __init__(self):
        self.id = np.arange(25)
        self.array = np.arange(25)
        self.array_opt = np.arange(25)
        self.list = np.arange(25).tolist()
        self.sparse_arr = sparse.csr.csr_matrix(np.zeros((25, 2)))
        self.name = 'name class'

class TestChecks(unittest.TestCase):
    """Test loading funcions from the Hazard class"""

    def test_check_oligatories_pass(self):
        """Correct DummyClass definition"""
        dummy = DummyClass()
        u_check.check_oligatories(dummy.__dict__, dummy.vars_oblig, "DummyClass.",
                                  dummy.id.size, dummy.id.size, 2)

    def test_check_oligatories_fail(self):
        """Wrong DummyClass definition"""
        dummy = DummyClass()
        dummy.array = np.arange(3)
        with self.assertRaises(ValueError) as cm:
            u_check.check_oligatories(dummy.__dict__, dummy.vars_oblig, "DummyClass.",
                                      dummy.id.size, dummy.id.size, 2)
        self.assertIn('Invalid DummyClass.array size: 25 != 3.', str(cm.exception))

        dummy = DummyClass()
        dummy.sparse_arr = sparse.csr.csr_matrix(np.zeros((25, 1)))
        with self.assertRaises(ValueError) as cm:
            u_check.check_oligatories(dummy.__dict__, dummy.vars_oblig, "DummyClass.",
                                      dummy.id.size, dummy.id.size, 2)
        self.assertIn('Invalid DummyClass.sparse_arr column size: 2 != 1.', str(cm.exception))

    def test_check_optionals_pass(self):
        """Correct DummyClass definition"""
        dummy = DummyClass()
        u_check.check_optionals(dummy.__dict__, dummy.vars_opt, "DummyClass.", dummy.id.size)

    def test_check_optionals_fail(self):
        """Correct DummyClass definition"""
        dummy = DummyClass()
        dummy.array_opt = np.arange(3)
        with self.assertRaises(ValueError) as cm:
            u_check.check_optionals(dummy.__dict__, dummy.vars_opt, "DummyClass.", dummy.id.size)
        self.assertIn('Invalid DummyClass.array_opt size: 25 != 3.', str(cm.exception))

        dummy.array_opt = np.array([], int)
        with self.assertLogs('climada.util.checker', level='DEBUG') as cm:
            u_check.check_optionals(dummy.__dict__, dummy.vars_opt, "DummyClass.", dummy.id.size)
        self.assertIn('DummyClass.array_opt not set.', cm.output[0])

        dummy = DummyClass()
        dummy.list = np.arange(3).tolist()
        with self.assertRaises(ValueError) as cm:
            u_check.check_optionals(dummy.__dict__, dummy.vars_opt, "DummyClass.", dummy.id.size)
        self.assertIn('Invalid DummyClass.list size: 25 != 3.', str(cm.exception))

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestChecks)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Hazard base class.
"""

import unittest
import datetime as dt
from pathlib import Path
import numpy as np
from scipy import sparse
from pathos.pools import ProcessPool as Pool

from climada import CONFIG
from climada.hazard.base import Hazard
from climada.hazard.centroids.centr import Centroids
import climada.util.dates_times as u_dt
from climada.util.constants import HAZ_TEMPLATE_XLS, HAZ_DEMO_FL
import climada.util.coordinates as u_coord
import climada.hazard.test as hazard_test

DATA_DIR = CONFIG.hazard.test_data.dir()
HAZ_TEST_MAT = Path(hazard_test.__file__).parent.joinpath('data', 'atl_prob_no_name.mat')

def dummy_hazard():
    hazard = Hazard('TC')
    hazard.tag.file_name = 'file1.mat'
    hazard.tag.description = 'Description 1'
    hazard.centroids = Centroids.from_lat_lon(np.array([1, 3, 5]), np.array([2, 4, 6]))
    hazard.event_id = np.array([1, 2, 3, 4])
    hazard.event_name = ['ev1', 'ev2', 'ev3', 'ev4']
    hazard.date = np.array([1, 2, 3, 4])
    hazard.orig = np.array([True, False, False, True])
    hazard.frequency = np.array([0.1, 0.5, 0.5, 0.2])
    hazard.fraction = sparse.csr_matrix([[0.02, 0.03, 0.04],
                                         [0.01, 0.01, 0.01],
                                         [0.3, 0.1, 0.0],
                                         [0.3, 0.2, 0.0]])
    hazard.intensity = sparse.csr_matrix([[0.2, 0.3, 0.4],
                                          [0.1, 0.1, 0.01],
                                          [4.3, 2.1, 1.0],
                                          [5.3, 0.2, 1.3]])
    hazard.units = 'm/s'

    return hazard

class TestLoader(unittest.TestCase):
    """Test loading funcions from the Hazard class"""

    @staticmethod
    def good_hazard():
        """Define well a hazard"""
        haz = Hazard('TC')
        haz.centroids = Centroids.from_lat_lon(np.array([1, 3]), np.array([2, 3]))
        haz.centroids.region_id = np.array([1, 2])
        haz.event_id = np.array([1, 2, 3])
        haz.event_name = ['A', 'B', 'C']
        haz.frequency = np.array([1, 2, 3])
        # events x centroids
        haz.intensity = sparse.csr_matrix([[1, 2], [1, 2], [1, 2]])
        haz.fraction = sparse.csr_matrix([[1, 2], [1, 2], [1, 2]])

        return haz

    def test_check_wrongCentroids_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.centroids.region_id = np.array([1, 2, 3, 4])
        with self.assertRaises(ValueError):
            haz.check()

    def test_check_wrongFreq_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.frequency = np.array([1, 2])

        with self.assertRaises(ValueError) as cm:
            haz.check()
        self.assertIn('Invalid Hazard.frequency size: 3 != 2.', str(cm.exception))

    def test_check_wrongInten_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.intensity = sparse.csr_matrix([[1, 2], [1, 2]])

        with self.assertRaises(ValueError) as cm:
            haz.check()
        self.assertIn('Invalid Hazard.intensity row size: 3 != 2.', str(cm.exception))

    def test_check_wrongFrac_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.fraction = sparse.csr_matrix([[1], [1], [1]])

        with self.assertRaises(ValueError) as cm:
            haz.check()
        self.assertIn('Invalid Hazard.fraction column size: 2 != 1.', str(cm.exception))

    def test_check_wrongEvName_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.event_name = ['M']

        with self.assertRaises(ValueError) as cm:
            haz.check()
        self.assertIn('Invalid Hazard.event_name size: 3 != 1.', str(cm.exception))

    def test_check_wrongId_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.event_id = np.array([1, 2, 1])

        with self.assertRaises(ValueError) as cm:
            haz.check()
        self.assertIn('There are events with the same identifier.', str(cm.exception))

    def test_check_wrong_date_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.date = np.array([1, 2])

        with self.assertRaises(ValueError) as cm:
            haz.check()
        self.assertIn('Invalid Hazard.date size: 3 != 2.', str(cm.exception))

    def test_check_wrong_orig_fail(self):
        """Wrong hazard definition"""
        haz = self.good_hazard()
        haz.orig = np.array([1, 2, 3, 4])

        with self.assertRaises(ValueError) as cm:
            haz.check()
        self.assertIn('Invalid Hazard.orig size: 3 != 4.', str(cm.exception))

    def test_event_name_to_id_pass(self):
        """Test event_name_to_id function."""
        haz = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        self.assertEqual(haz.get_event_id('event001')[0], 1)
        self.assertEqual(haz.get_event_id('event084')[0], 84)

    def test_event_name_to_id_fail(self):
        """Test event_name_to_id function."""
        haz = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        with self.assertRaises(ValueError) as cm:
            haz.get_event_id('1050')
        self.assertIn('No event with name: 1050', str(cm.exception))

    def test_event_id_to_name_pass(self):
        """Test event_id_to_name function."""
        haz = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        self.assertEqual(haz.get_event_name(2), 'event002')
        self.assertEqual(haz.get_event_name(48), 'event048')

    def test_event_id_to_name_fail(self):
        """Test event_id_to_name function."""
        haz = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        with self.assertRaises(ValueError) as cm:
            haz.get_event_name(1050)
        self.assertIn('No event with id: 1050', str(cm.exception))

    def test_get_date_strings_pass(self):
        haz = Hazard.from_mat(HAZ_TEST_MAT)
        haz.event_name[5] = 'HAZEL'
        haz.event_name[10] = 'HAZEL'

        self.assertEqual(len(haz.get_event_date('HAZEL')), 2)
        self.assertEqual(haz.get_event_date('HAZEL')[0],
                         u_dt.date_to_str(haz.date[5]))
        self.assertEqual(haz.get_event_date('HAZEL')[1],
                         u_dt.date_to_str(haz.date[10]))

        self.assertEqual(haz.get_event_date(2)[0], u_dt.date_to_str(haz.date[1]))

        self.assertEqual(len(haz.get_event_date()), haz.date.size)
        self.assertEqual(haz.get_event_date()[560],
                         u_dt.date_to_str(haz.date[560]))

class TestRemoveDupl(unittest.TestCase):
    """Test remove_duplicates method."""

    def test_equal_same(self):
        """Append the same hazard and remove duplicates, obtain initial hazard."""
        haz1 = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        haz2 = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        haz1.append(haz2)
        haz1.remove_duplicates()
        haz1.check()
        self.assertEqual(haz1.event_name, haz2.event_name)
        self.assertTrue(np.array_equal(haz1.event_id, haz2.event_id))
        self.assertTrue(np.array_equal(haz1.frequency, haz2.frequency))
        self.assertTrue(np.array_equal(haz1.date, haz2.date))
        self.assertTrue(np.array_equal(haz1.orig, haz2.orig))
        self.assertTrue(np.array_equal(haz1.intensity.toarray(), haz2.intensity.toarray()))
        self.assertTrue(np.array_equal(haz1.fraction.toarray(), haz2.fraction.toarray()))
        self.assertTrue((haz1.intensity != haz2.intensity).nnz == 0)
        self.assertTrue((haz1.fraction != haz2.fraction).nnz == 0)
        self.assertEqual(haz1.units, haz2.units)
        self.assertEqual(haz1.tag.file_name, [haz2.tag.file_name, haz2.tag.file_name])
        self.assertEqual(haz1.tag.haz_type, haz2.tag.haz_type)
        self.assertEqual(haz1.tag.description, [haz2.tag.description, haz2.tag.description])

    def test_same_events_same(self):
        """Append hazard with same events and diff centroids. After removing
        duplicate events, initial events are obtained with 0 intensity and
        fraction in new appended centroids."""
        haz1 = dummy_hazard()
        haz2 = Hazard('TC')
        haz2.tag.file_name = 'file2.mat',
        haz2.tag.description = 'Description 2'
        haz2.centroids = Centroids.from_lat_lon(np.array([7, 9, 11]), np.array([8, 10, 12]))

        haz2.event_id = haz1.event_id
        haz2.event_name = haz1.event_name
        haz2.frequency = haz1.frequency
        haz2.date = haz1.date
        haz2.fraction = sparse.csr_matrix([[0.22, 0.32, 0.44],
                                           [0.11, 0.11, 0.11],
                                           [0.32, 0.11, 0.99],
                                           [0.32, 0.22, 0.88]])
        haz2.intensity = sparse.csr_matrix([[0.22, 3.33, 6.44],
                                            [1.11, 0.11, 1.11],
                                            [8.33, 4.11, 4.4],
                                            [9.33, 9.22, 1.77]])
        haz2.units = 'm/s'

        haz1.append(haz2)
        haz1.remove_duplicates()
        haz1.check()

        # expected values
        haz_res = dummy_hazard()
        haz_res.intensity = sparse.hstack(
            [haz_res.intensity, sparse.csr_matrix((haz_res.intensity.shape[0], 3))], format='csr')
        haz_res.fraction = sparse.hstack(
            [haz_res.fraction, sparse.csr_matrix((haz_res.fraction.shape[0], 3))], format='csr')
        self.assertTrue(np.array_equal(haz_res.intensity.toarray(),
                                       haz1.intensity.toarray()))
        self.assertTrue(sparse.isspmatrix_csr(haz1.intensity))
        self.assertTrue(np.array_equal(haz_res.fraction.toarray(),
                                       haz1.fraction.toarray()))
        self.assertTrue(sparse.isspmatrix_csr(haz1.fraction))
        self.assertEqual(haz1.event_name, haz_res.event_name)
        self.assertTrue(np.array_equal(haz1.date, haz_res.date))
        self.assertTrue(np.array_equal(haz1.orig, haz_res.orig))
        self.assertTrue(np.array_equal(haz1.event_id,
                                       haz_res.event_id))
        self.assertTrue(np.array_equal(haz1.frequency, haz_res.frequency))
        self.assertEqual(haz_res.units, haz1.units)

        self.assertEqual(haz1.tag.file_name,
                         [haz_res.tag.file_name, haz2.tag.file_name])
        self.assertEqual(haz1.tag.haz_type, haz_res.tag.haz_type)
        self.assertEqual(haz1.tag.description,
                         [haz_res.tag.description, haz2.tag.description])

class TestSelect(unittest.TestCase):
    """Test select method."""

    def test_select_event_name(self):
        """Test select historical events."""
        haz = dummy_hazard()
        sel_haz = haz.select(event_names=['ev4', 'ev1'])

        self.assertTrue(np.array_equal(sel_haz.centroids.coord, haz.centroids.coord))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, np.array([4, 1])))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([4, 1])))
        self.assertTrue(np.array_equal(sel_haz.orig, np.array([True, True])))
        self.assertTrue(np.array_equal(sel_haz.frequency, np.array([0.2, 0.1])))
        self.assertTrue(np.array_equal(sel_haz.fraction.toarray(),
                                       np.array([[0.3, 0.2, 0.0],
                                                 [0.02, 0.03, 0.04]])))
        self.assertTrue(np.array_equal(sel_haz.intensity.toarray(),
                                       np.array([[5.3, 0.2, 1.3],
                                                 [0.2, 0.3, 0.4]])))
        self.assertEqual(sel_haz.event_name, ['ev4', 'ev1'])
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

    def test_select_orig_pass(self):
        """Test select historical events."""
        haz = dummy_hazard()
        sel_haz = haz.select(orig=True)

        self.assertTrue(np.array_equal(sel_haz.centroids.coord, haz.centroids.coord))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, np.array([1, 4])))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([1, 4])))
        self.assertTrue(np.array_equal(sel_haz.orig, np.array([True, True])))
        self.assertTrue(np.array_equal(sel_haz.frequency, np.array([0.1, 0.2])))
        self.assertTrue(np.array_equal(
            sel_haz.fraction.toarray(), np.array([[0.02, 0.03, 0.04], [0.3, 0.2, 0.0]])))
        self.assertTrue(np.array_equal(
            sel_haz.intensity.toarray(), np.array([[0.2, 0.3, 0.4], [5.3, 0.2, 1.3]])))
        self.assertEqual(sel_haz.event_name, ['ev1', 'ev4'])
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

    def test_select_syn_pass(self):
        """Test select historical events."""
        haz = dummy_hazard()
        sel_haz = haz.select(orig=False)

        self.assertTrue(np.array_equal(sel_haz.centroids.coord, haz.centroids.coord))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.orig, np.array([False, False])))
        self.assertTrue(np.array_equal(sel_haz.frequency, np.array([0.5, 0.5])))
        self.assertTrue(np.array_equal(
            sel_haz.fraction.toarray(), np.array([[0.01, 0.01, 0.01], [0.3, 0.1, 0.0]])))
        self.assertTrue(np.array_equal(
            sel_haz.intensity.toarray(), np.array([[0.1, 0.1, 0.01], [4.3, 2.1, 1.0]])))
        self.assertEqual(sel_haz.event_name, ['ev2', 'ev3'])
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

    def test_select_date_pass(self):
        """Test select historical events."""
        haz = dummy_hazard()
        sel_haz = haz.select(date=(2, 4))

        self.assertTrue(np.array_equal(sel_haz.centroids.coord, haz.centroids.coord))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, np.array([2, 3, 4])))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([2, 3, 4])))
        self.assertTrue(np.array_equal(sel_haz.orig, np.array([False, False, True])))
        self.assertTrue(np.array_equal(sel_haz.frequency, np.array([0.5, 0.5, 0.2])))
        self.assertTrue(np.array_equal(
            sel_haz.fraction.toarray(), np.array([[0.01, 0.01, 0.01],
                                                  [0.3, 0.1, 0.0],
                                                  [0.3, 0.2, 0.0]])))
        self.assertTrue(np.array_equal(
            sel_haz.intensity.toarray(), np.array([[0.1, 0.1, 0.01],
                                                   [4.3, 2.1, 1.0],
                                                   [5.3, 0.2, 1.3]])))
        self.assertEqual(sel_haz.event_name, ['ev2', 'ev3', 'ev4'])
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

    def test_select_date_str_pass(self):
        """Test select historical events."""
        haz = dummy_hazard()
        sel_haz = haz.select(date=('0001-01-02', '0001-01-03'))

        self.assertTrue(np.array_equal(sel_haz.centroids.coord, haz.centroids.coord))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.orig, np.array([False, False])))
        self.assertTrue(np.array_equal(sel_haz.frequency, np.array([0.5, 0.5])))
        self.assertTrue(np.array_equal(
            sel_haz.fraction.toarray(), np.array([[0.01, 0.01, 0.01], [0.3, 0.1, 0.0]])))
        self.assertTrue(np.array_equal(
            sel_haz.intensity.toarray(), np.array([[0.1, 0.1, 0.01], [4.3, 2.1, 1.0]])))
        self.assertEqual(sel_haz.event_name, ['ev2', 'ev3'])
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

    def test_select_date_and_orig_pass(self):
        """Test select historical events."""
        haz = dummy_hazard()
        sel_haz = haz.select(date=(2, 4), orig=False)

        self.assertTrue(np.array_equal(sel_haz.centroids.coord, haz.centroids.coord))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.orig, np.array([False, False])))
        self.assertTrue(np.array_equal(sel_haz.frequency, np.array([0.5, 0.5])))
        self.assertTrue(np.array_equal(
            sel_haz.fraction.toarray(), np.array([[0.01, 0.01, 0.01], [0.3, 0.1, 0.0]])))
        self.assertTrue(np.array_equal(
            sel_haz.intensity.toarray(), np.array([[0.1, 0.1, 0.01], [4.3, 2.1, 1.0]])))
        self.assertEqual(sel_haz.event_name, ['ev2', 'ev3'])
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

    def test_select_date_wrong_pass(self):
        """Test select historical events."""
        haz = dummy_hazard()
        sel_haz = haz.select(date=(6, 8), orig=False)
        self.assertEqual(sel_haz, None)

    def test_select_date_invalid_pass(self):
        """Test select with invalid date values"""
        haz = dummy_hazard()

        # lists and numpy arrays should work just like tuples
        sel_haz = haz.select(date=['0001-01-02', '0001-01-03'])
        self.assertTrue(np.array_equal(sel_haz.date, np.array([2, 3])))
        sel_haz = haz.select(date=np.array([2, 4]))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([2, 3, 4])))

        # iterables of length not exactly 2 are invalid
        with self.assertRaises(ValueError) as context:
            haz.select(date=(3,))
        self.assertTrue("not enough values to unpack" in str(context.exception))
        with self.assertRaises(ValueError) as context:
            haz.select(date=(1, 2, 3))
        self.assertTrue("too many values to unpack" in str(context.exception))

        # only strings and numbers are valid
        with self.assertRaises(TypeError) as context:
            haz.select(date=({}, {}))

    def test_select_reg_id_pass(self):
        """Test select region of centroids."""
        haz = dummy_hazard()
        haz.centroids.region_id = np.array([5, 7, 9])
        sel_haz = haz.select(date=(2, 4), orig=False, reg_id=9)

        self.assertTrue(np.array_equal(sel_haz.centroids.coord.squeeze(),
                                       haz.centroids.coord[2, :]))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.date, np.array([2, 3])))
        self.assertTrue(np.array_equal(sel_haz.orig, np.array([False, False])))
        self.assertTrue(np.array_equal(sel_haz.frequency, np.array([0.5, 0.5])))
        self.assertTrue(np.array_equal(sel_haz.fraction.toarray(), np.array([[0.01], [0.0]])))
        self.assertTrue(np.array_equal(sel_haz.intensity.toarray(), np.array([[0.01], [1.0]])))
        self.assertEqual(sel_haz.event_name, ['ev2', 'ev3'])
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

    def test_select_tight_pass(self):
        """Test select tight box around hazard"""

        #intensity select
        haz = dummy_hazard()
        haz.intensity[:, -1] = 0.0
        sel_haz = haz.select_tight()

        self.assertTrue(np.array_equal(sel_haz.centroids.coord.squeeze(),
                                       haz.centroids.coord[:-1, :]))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, haz.event_id))
        self.assertTrue(np.array_equal(sel_haz.date, haz.date))
        self.assertTrue(np.array_equal(sel_haz.orig, haz.orig))
        self.assertTrue(np.array_equal(sel_haz.frequency, haz.frequency))
        self.assertTrue(np.array_equal(sel_haz.fraction.toarray(),
                                       haz.fraction[:,:-1].toarray()))
        self.assertTrue(np.array_equal(sel_haz.intensity.toarray(),
                                       haz.intensity[:,:-1].toarray()))
        self.assertEqual(sel_haz.event_name, haz.event_name)
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)

        #fraction select
        haz = dummy_hazard()
        haz.fraction[:, -1] = 0.0

        sel_haz = haz.select_tight(val='fraction')

        self.assertTrue(np.array_equal(sel_haz.centroids.coord.squeeze(),
                                       haz.centroids.coord[:-1, :]))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, haz.event_id))
        self.assertTrue(np.array_equal(sel_haz.date, haz.date))
        self.assertTrue(np.array_equal(sel_haz.orig, haz.orig))
        self.assertTrue(np.array_equal(sel_haz.frequency, haz.frequency))
        self.assertTrue(np.array_equal(sel_haz.fraction.toarray(),
                                       haz.fraction[:,:-1].toarray()))
        self.assertTrue(np.array_equal(sel_haz.intensity.toarray(),
                                       haz.intensity[:,:-1].toarray()))
        self.assertEqual(sel_haz.event_name, haz.event_name)
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)


        haz = dummy_hazard()
        haz.intensity[:, -1] = 0.0

        # small buffer: zero field is discarded
        sel_haz = haz.select_tight(buffer=0.1)
        self.assertTrue(np.array_equal(sel_haz.centroids.coord.squeeze(),
                                       haz.centroids.coord[:-1, :]))
        # large buffer: zero field is retained
        sel_haz = haz.select_tight(buffer=10)

        self.assertTrue(np.array_equal(sel_haz.centroids.coord.squeeze(),
                                       haz.centroids.coord))
        self.assertEqual(sel_haz.tag, haz.tag)
        self.assertEqual(sel_haz.units, haz.units)
        self.assertTrue(np.array_equal(sel_haz.event_id, haz.event_id))
        self.assertTrue(np.array_equal(sel_haz.date, haz.date))
        self.assertTrue(np.array_equal(sel_haz.orig, haz.orig))
        self.assertTrue(np.array_equal(sel_haz.frequency, haz.frequency))
        self.assertTrue(np.array_equal(sel_haz.fraction.toarray(),
                                       haz.fraction.toarray()))
        self.assertTrue(np.array_equal(sel_haz.intensity.toarray(),
                                       haz.intensity.toarray()))
        self.assertEqual(sel_haz.event_name, haz.event_name)
        self.assertIsInstance(sel_haz, Hazard)
        self.assertIsInstance(sel_haz.intensity, sparse.csr_matrix)
        self.assertIsInstance(sel_haz.fraction, sparse.csr_matrix)


class TestAppend(unittest.TestCase):
    """Test append method."""

    def test_append_empty_fill(self):
        """Append an empty. Obtain initial hazard."""
        haz1 = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        haz2 = Hazard('TC')
        haz2.centroids.geometry.crs = 'epsg:4326'
        haz1.append(haz2)
        haz1.check()

        # expected values
        haz1_orig = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        self.assertEqual(haz1.event_name, haz1_orig.event_name)
        self.assertTrue(np.array_equal(haz1.event_id, haz1_orig.event_id))
        self.assertTrue(np.array_equal(haz1.date, haz1_orig.date))
        self.assertTrue(np.array_equal(haz1.orig, haz1_orig.orig))
        self.assertTrue(np.array_equal(haz1.frequency, haz1_orig.frequency))
        self.assertTrue((haz1.intensity != haz1_orig.intensity).nnz == 0)
        self.assertTrue((haz1.fraction != haz1_orig.fraction).nnz == 0)
        self.assertEqual(haz1.units, haz1_orig.units)
        self.assertEqual(haz1.tag.file_name, haz1_orig.tag.file_name)
        self.assertEqual(haz1.tag.haz_type, haz1_orig.tag.haz_type)
        self.assertEqual(haz1.tag.description, haz1_orig.tag.description)

    def test_append_to_empty_fill(self):
        """Append to an empty hazard a filled one. Obtain filled one."""
        haz1 = Hazard('TC')
        haz2 = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        haz1.append(haz2)
        haz1.check()

        # expected values
        haz1_orig = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        self.assertEqual(haz1.event_name, haz1_orig.event_name)
        self.assertTrue(np.array_equal(haz1.event_id, haz1_orig.event_id))
        self.assertTrue(np.array_equal(haz1.frequency, haz1_orig.frequency))
        self.assertTrue(np.array_equal(haz1.date, haz1_orig.date))
        self.assertTrue(np.array_equal(haz1.orig, haz1_orig.orig))
        self.assertTrue((haz1.intensity != haz1_orig.intensity).nnz == 0)
        self.assertTrue((haz1.fraction != haz1_orig.fraction).nnz == 0)
        self.assertEqual(haz1.units, haz1_orig.units)
        self.assertEqual(haz1.tag.file_name, haz1_orig.tag.file_name)
        self.assertEqual(haz1.tag.haz_type, haz1_orig.tag.haz_type)
        self.assertEqual(haz1.tag.description, haz1_orig.tag.description)

    def test_same_centroids_extend(self):
        """Append hazard with same centroids, different events."""
        haz1 = dummy_hazard()
        haz2 = Hazard('TC')
        haz2.tag.file_name = 'file2.mat'
        haz2.tag.description = 'Description 2'

        haz2.centroids = haz1.centroids
        haz2.event_id = np.array([5, 6, 7, 8])
        haz2.event_name = ['ev5', 'ev6', 'ev7', 'ev8']
        haz2.frequency = np.array([0.9, 0.75, 0.75, 0.22])
        haz2.fraction = sparse.csr_matrix([[0.2, 0.3, 0.4],
                                           [0.1, 0.1, 0.1],
                                           [0.3, 0.1, 0.9],
                                           [0.3, 0.2, 0.8]])
        haz2.intensity = sparse.csr_matrix([[0.2, 3.3, 6.4],
                                            [1.1, 0.1, 1.01],
                                            [8.3, 4.1, 4.0],
                                            [9.3, 9.2, 1.7]])
        haz2.units = 'm/s'

        haz1.append(haz2)
        haz1.check()

        # expected values
        haz1_orig = dummy_hazard()
        exp_inten = np.zeros((8, 3))
        exp_inten[0:4, 0:3] = haz1_orig.intensity.toarray()
        exp_inten[4:8, 0:3] = haz2.intensity.toarray()
        exp_frac = np.zeros((8, 3))
        exp_frac[0:4, 0:3] = haz1_orig.fraction.toarray()
        exp_frac[4:8, 0:3] = haz2.fraction.toarray()

        self.assertEqual(haz1.event_id.size, 8)
        self.assertTrue(sparse.isspmatrix_csr(haz1.intensity))
        self.assertTrue(sparse.isspmatrix_csr(haz1.fraction))
        for i_ev in range(haz1.event_id.size):
            self.assertTrue(any((haz1.intensity[i_ev].toarray() == exp_inten).all(1)))
            self.assertTrue(any((haz1.fraction[i_ev].toarray() == exp_frac).all(1)))
            self.assertTrue(haz1.event_name[i_ev] in haz1_orig.event_name + haz2.event_name)
            self.assertTrue(haz1.date[i_ev] in np.append(haz1_orig.date, haz2.date))
            self.assertTrue(haz1.orig[i_ev] in np.append(haz1_orig.orig, haz2.orig))
            self.assertTrue(haz1.event_id[i_ev] in np.append(haz1_orig.event_id, haz2.event_id))
            self.assertTrue(haz1.frequency[i_ev] in np.append(haz1_orig.frequency, haz2.frequency))

        self.assertEqual(haz1.centroids.size, 3)
        self.assertTrue(np.array_equal(haz1.centroids.coord, haz2.centroids.coord))
        self.assertEqual(haz1.tag.file_name,
                         [haz1_orig.tag.file_name, haz2.tag.file_name])
        self.assertEqual(haz1.tag.haz_type, haz1_orig.tag.haz_type)
        self.assertEqual(haz1.tag.description,
                         [haz1_orig.tag.description, haz2.tag.description])

    def test_incompatible_type_fail(self):
        """Raise error when append two incompatible hazards."""
        haz1 = dummy_hazard()
        haz2 = dummy_hazard()
        haz2.tag.haz_type = 'WS'
        haz2.tag.file_name = 'file2.mat'
        haz2.tag.description = 'Description 2'
        with self.assertRaises(ValueError) as cm:
            haz1.append(haz2)

    def test_incompatible_units_fail(self):
        """Raise error when append two incompatible hazards."""
        haz1 = dummy_hazard()
        haz2 = dummy_hazard()
        haz2.units = 'km/h'
        with self.assertRaises(ValueError) as cm:
            haz1.append(haz2)

    def test_all_different_extend(self):
        """Append totally different hazard."""
        haz1 = dummy_hazard()
        haz2 = Hazard('TC')
        haz2.tag.file_name = 'file2.mat'
        haz2.tag.description = 'Description 2'
        haz2.centroids = Centroids.from_lat_lon(np.array([7, 9, 11]), np.array([8, 10, 12]))
        haz2.event_id = np.array([5, 6, 7, 8])
        haz2.event_name = ['ev5', 'ev6', 'ev7', 'ev8']
        haz2.frequency = np.array([0.9, 0.75, 0.75, 0.22])
        haz2.fraction = sparse.csr_matrix([[0.2, 0.3, 0.4],
                                           [0.1, 0.1, 0.1],
                                           [0.3, 0.1, 0.9],
                                           [0.3, 0.2, 0.8]])
        haz2.intensity = sparse.csr_matrix([[0.2, 3.3, 6.4],
                                            [1.1, 0.1, 1.01],
                                            [8.3, 4.1, 4.0],
                                            [9.3, 9.2, 1.7]])
        haz2.date = np.ones((4,))
        haz2.orig = np.ones((4,))
        haz2.units = 'm/s'

        haz1.append(haz2)
        haz1.check()

        # expected values
        haz1_orig = dummy_hazard()
        exp_inten = np.zeros((8, 6))
        exp_inten[0:4, 0:3] = haz1_orig.intensity.toarray()
        exp_inten[4:8, 3:6] = haz2.intensity.toarray()
        exp_frac = np.zeros((8, 6))
        exp_frac[0:4, 0:3] = haz1_orig.fraction.toarray()
        exp_frac[4:8, 3:6] = haz2.fraction.toarray()
        self.assertEqual(haz1.event_id.size, 8)
        self.assertTrue(sparse.isspmatrix_csr(haz1.intensity))
        self.assertTrue(sparse.isspmatrix_csr(haz1.fraction))
        for i_ev in range(haz1.event_id.size):
            self.assertTrue(any((haz1.intensity[i_ev].toarray() == exp_inten).all(1)))
            self.assertTrue(any((haz1.fraction[i_ev].toarray() == exp_frac).all(1)))
            self.assertTrue(haz1.event_name[i_ev] in haz1_orig.event_name + haz2.event_name)
            self.assertTrue(haz1.date[i_ev] in np.append(haz1_orig.date, haz2.date))
            self.assertTrue(haz1.orig[i_ev] in np.append(haz1_orig.orig, haz2.orig))
            self.assertTrue(haz1.event_id[i_ev] in np.append(haz1_orig.event_id, haz2.event_id))
            self.assertTrue(haz1.frequency[i_ev] in np.append(haz1_orig.frequency, haz2.frequency))

        self.assertEqual(haz1.centroids.size, 6)
        self.assertEqual(haz1_orig.units, haz1.units)
        self.assertEqual(haz1.tag.file_name,
                         [haz1_orig.tag.file_name, haz2.tag.file_name])
        self.assertEqual(haz1.tag.haz_type, haz1_orig.tag.haz_type)
        self.assertEqual(haz1.tag.description,
                         [haz1_orig.tag.description, haz2.tag.description])

    def test_same_events_append(self):
        """Append hazard with same events (and diff centroids).
        Events are appended with all new centroids columns."""
        haz1 = dummy_hazard()
        haz2 = Hazard('TC')
        haz2.tag.file_name = 'file2.mat'
        haz2.tag.description = 'Description 2'
        haz2.centroids = Centroids.from_lat_lon(np.array([7, 9, 11]), np.array([8, 10, 12]))

        haz2.event_id = haz1.event_id
        haz2.event_name = haz1.event_name.copy()
        haz2.frequency = haz1.frequency
        haz2.date = haz1.date
        haz2.fraction = sparse.csr_matrix([[0.22, 0.32, 0.44],
                                           [0.11, 0.11, 0.11],
                                           [0.32, 0.11, 0.99],
                                           [0.32, 0.22, 0.88]])
        haz2.intensity = sparse.csr_matrix([[0.22, 3.33, 6.44],
                                            [1.11, 0.11, 1.11],
                                            [8.33, 4.11, 4.4],
                                            [9.33, 9.22, 1.77]])
        haz2.units = 'm/s'

        haz1.append(haz2)

        # expected values
        haz1_ori = dummy_hazard()
        res_inten = np.zeros((8, 6))
        res_inten[0:4, 0:3] = haz1_ori.intensity.toarray()
        res_inten[4:, 3:] = haz2.intensity.toarray()

        res_frac = np.zeros((8, 6))
        res_frac[0:4, 0:3] = haz1_ori.fraction.toarray()
        res_frac[4:, 3:] = haz2.fraction.toarray()

        self.assertTrue(np.array_equal(res_inten,
                                       haz1.intensity.toarray()))
        self.assertTrue(sparse.isspmatrix_csr(haz1.intensity))
        self.assertTrue(np.array_equal(res_frac,
                                       haz1.fraction.toarray()))
        self.assertTrue(sparse.isspmatrix_csr(haz1.fraction))
        self.assertEqual(haz1.event_name,
                         haz1_ori.event_name + haz2.event_name)
        self.assertTrue(np.array_equal(haz1.date,
                                       np.append(haz1_ori.date, haz2.date)))
        self.assertTrue(np.array_equal(haz1.orig,
                                       np.append(haz1_ori.orig, haz2.orig)))
        self.assertTrue(np.array_equal(haz1.event_id, np.arange(1, 9)))
        self.assertTrue(np.array_equal(haz1.frequency,
                                       np.append(haz1_ori.frequency, haz2.frequency)))
        self.assertEqual(haz1_ori.units, haz1.units)

        self.assertEqual(haz1.tag.file_name,
                         [haz1_ori.tag.file_name, haz2.tag.file_name])
        self.assertEqual(haz1.tag.haz_type, haz1_ori.tag.haz_type)
        self.assertEqual(haz1.tag.description,
                         [haz1_ori.tag.description, haz2.tag.description])

    def test_concat_pass(self):
        """Test concatenate function."""
        class TCHazard(Hazard):
            # derived class of Hazard for testing purposes
            def __init__(self):
                # only define hazard type
                Hazard.__init__(self, 'TC')

        haz_1 = TCHazard()
        haz_1.tag.file_name = 'file1.mat'
        haz_1.tag.description = 'Description 1'
        haz_1.centroids = Centroids.from_lat_lon(np.array([1, 3, 5]), np.array([2, 4, 6]))
        haz_1.event_id = np.array([1])
        haz_1.event_name = ['ev1']
        haz_1.date = np.array([1])
        haz_1.orig = np.array([True])
        haz_1.frequency = np.array([1.0])
        haz_1.fraction = sparse.csr_matrix([[0.02, 0.03, 0.04]])
        haz_1.intensity = sparse.csr_matrix([[0.2, 0.3, 0.4]])
        haz_1.units = 'm/s'

        haz_2 = TCHazard()
        haz_2.tag.file_name = 'file2.mat'
        haz_2.tag.description = 'Description 2'
        haz_2.centroids = Centroids.from_lat_lon(np.array([1, 3, 5]), np.array([2, 4, 6]))
        haz_2.event_id = np.array([1])
        haz_2.event_name = ['ev2']
        haz_2.date = np.array([2])
        haz_2.orig = np.array([False])
        haz_2.frequency = np.array([1.0])
        haz_2.fraction = sparse.csr_matrix([[1.02, 1.03, 1.04]])
        haz_2.intensity = sparse.csr_matrix([[1.2, 1.3, 1.4]])
        haz_2.units = 'm/s'

        haz = TCHazard.concat([haz_1, haz_2])

        hres_frac = sparse.csr_matrix([[0.02, 0.03, 0.04],
                                       [1.02, 1.03, 1.04]])
        hres_inten = sparse.csr_matrix([[0.2, 0.3, 0.4],
                                        [1.2, 1.3, 1.4]])

        self.assertIsInstance(haz, TCHazard)
        self.assertTrue(sparse.isspmatrix_csr(haz.intensity))
        self.assertTrue(np.array_equal(haz.intensity.toarray(), hres_inten.toarray()))
        self.assertTrue(sparse.isspmatrix_csr(haz.fraction))
        self.assertTrue(np.array_equal(haz.fraction.toarray(), hres_frac.toarray()))
        self.assertEqual(haz.units, haz_2.units)
        self.assertTrue(np.array_equal(haz.frequency, np.array([1.0, 1.0])))
        self.assertTrue(np.array_equal(haz.orig, np.array([True, False])))
        self.assertTrue(np.array_equal(haz.date, np.array([1, 2])))
        self.assertTrue(np.array_equal(haz.event_id, np.array([1, 2])))
        self.assertEqual(haz.event_name, ['ev1', 'ev2'])
        self.assertTrue(np.array_equal(haz.centroids.coord, haz_1.centroids.coord))
        self.assertTrue(np.array_equal(haz.centroids.coord, haz_2.centroids.coord))
        self.assertEqual(haz.tag.file_name, ['file1.mat', 'file2.mat'])
        self.assertEqual(haz.tag.description, ['Description 1', 'Description 2'])

    def test_append_new_var_pass(self):
        """New variable appears if hazard to append is empty."""
        haz = dummy_hazard()
        haz.new_var = np.ones(haz.size)

        app_haz = Hazard('TC')
        app_haz.append(haz)
        self.assertIn('new_var', app_haz.__dict__)

    def test_append_raise_type_error(self):
        """Raise error if hazards of different class"""
        haz1 = Hazard('TC')
        haz1.units = 'm/s'
        from climada.hazard import TropCyclone
        haz2 = TropCyclone()
        with self.assertRaises(TypeError):
            haz1.append(haz2)

    def test_concat_raise_value_error(self):
        """Raise error if hazards with different units of type"""

        haz1 = Hazard('TC')
        haz1.units = 'm/s'
        haz3 = Hazard('EQ')
        with self.assertRaises(ValueError):
             Hazard.concat([haz1, haz3])

        haz4 = Hazard('TC')
        haz4.units = 'cm'
        with self.assertRaises(ValueError):
             Hazard.concat([haz1, haz4])

    def test_change_centroids(self):
        """Set new centroids for hazard"""
        cent1 = Centroids()
        cent1.lat, cent1.lon = np.array([0, 1]), np.array([0, -1])
        cent1.on_land = np.array([True, True])

        haz_1 = Hazard('TC')
        haz_1.tag.file_name = 'file1.mat'
        haz_1.tag.description = 'Description 1'
        haz_1.centroids = cent1
        haz_1.event_id = np.array([1])
        haz_1.event_name = ['ev1']
        haz_1.date = np.array([1])
        haz_1.orig = np.array([True])
        haz_1.frequency = np.array([1.0])
        haz_1.fraction = sparse.csr_matrix([[0.02, 0.03]])
        haz_1.intensity = sparse.csr_matrix([[0.2, 0.3]])
        haz_1.units = 'm/s'

        cent2 = Centroids()
        cent2.lat, cent2.lon = np.array([0, 1, 3]), np.array([0, -1, 3])
        cent2.on_land = np.array([True, True, False])

        haz_2 = haz_1.change_centroids(cent2)

        self.assertTrue(np.array_equal(haz_2.intensity.toarray(),
                               np.array([[0.2, 0.3, 0.]])))
        self.assertTrue(np.array_equal(haz_2.fraction.toarray(),
                               np.array([[0.02, 0.03, 0.]])))
        self.assertTrue(np.array_equal(haz_2.event_id, np.array([1])))
        self.assertTrue(np.array_equal(haz_2.event_name, ['ev1']))
        self.assertTrue(np.array_equal(haz_2.orig, [True]))
        self.assertEqual(haz_2.tag.description, 'Description 1')

        """Test error for projection"""
        cent3 = Centroids()
        cent3.lat, cent3.lon = np.array([0.5, 3]), np.array([-0.5, 3])
        cent3.on_land = np.array([True, True, False])

        with self.assertRaises(ValueError) as cm:
            haz_1.change_centroids(cent3, threshold=100)
        self.assertIn('two hazard centroids are mapped to the same centroids', str(cm.exception))


    def test_change_centroids_raster(self):
        """Set new centroids for hazard"""
        cent1 = Centroids()
        cent1.lat, cent1.lon = np.array([0, 1]), np.array([0, -1])
        cent1.on_land = np.array([True, True])

        haz_1 = Hazard('TC')
        haz_1.tag.file_name = 'file1.mat'
        haz_1.tag.description = 'Description 1'
        haz_1.centroids = cent1
        haz_1.event_id = np.array([1])
        haz_1.event_name = ['ev1']
        haz_1.date = np.array([1])
        haz_1.orig = np.array([True])
        haz_1.frequency = np.array([1.0])
        haz_1.fraction = sparse.csr_matrix([[0.02, 0.03]])
        haz_1.intensity = sparse.csr_matrix([[0.2, 0.3]])
        haz_1.units = 'm/s'


        """Test with raster centroids"""
        cent4 = Centroids.from_pnt_bounds(points_bounds=(-1, 0, 0, 1), res=1)
        cent4.lat, cent1.lon = np.array([0, 1]), np.array([0, -1])
        cent4.on_land = np.array([True, True])

        haz_4 = haz_1.change_centroids(cent4)

        self.assertTrue(np.array_equal(haz_4.intensity.toarray(),
                               np.array([[0.3, 0.0, 0.0, 0.2]])))
        self.assertTrue(np.array_equal(haz_4.fraction.toarray(),
                               np.array([[0.03, 0.0, 0.0, 0.02]])))
        self.assertTrue(np.array_equal(haz_4.event_id, np.array([1])))
        self.assertTrue(np.array_equal(haz_4.event_name, ['ev1']))
        self.assertTrue(np.array_equal(haz_4.orig, [True]))
        self.assertEqual(haz_4.tag.description, 'Description 1')


class TestStats(unittest.TestCase):
    """Test return period statistics"""

    def test_degenerate_pass(self):
        """Test degenerate call."""
        haz = Hazard.from_mat(HAZ_TEST_MAT)
        return_period = np.array([25, 50, 100, 250])
        haz.intensity = sparse.csr.csr_matrix(np.zeros(haz.intensity.shape))
        inten_stats = haz.local_exceedance_inten(return_period)
        self.assertTrue(np.array_equal(inten_stats, np.zeros((4, 100))))

    def test_ref_all_pass(self):
        """Compare against reference."""
        haz = Hazard.from_mat(HAZ_TEST_MAT)
        return_period = np.array([25, 50, 100, 250])
        inten_stats = haz.local_exceedance_inten(return_period)

        self.assertAlmostEqual(inten_stats[0][0], 55.424015590131290)
        self.assertAlmostEqual(inten_stats[1][0], 67.221687644669998)
        self.assertAlmostEqual(inten_stats[2][0], 79.019359699208721)
        self.assertAlmostEqual(inten_stats[3][0], 94.615033842370963)

        self.assertAlmostEqual(inten_stats[1][66], 70.608592953031405)
        self.assertAlmostEqual(inten_stats[3][33], 88.510983305123631)
        self.assertAlmostEqual(inten_stats[2][99], 79.717518054203623)

class TestYearset(unittest.TestCase):
    """Test return period statistics"""

    def test_ref_pass(self):
        """Test against matlab reference."""
        haz = Hazard.from_mat(HAZ_TEST_MAT)
        orig_year_set = haz.calc_year_set()

        self.assertTrue(np.array_equal(np.array(list(orig_year_set.keys())),
                                       np.arange(1851, 2012)))
        self.assertTrue(np.array_equal(orig_year_set[1851],
                                       np.array([1, 11, 21, 31])))
        self.assertTrue(np.array_equal(orig_year_set[1958],
                                       np.array([8421, 8431, 8441, 8451, 8461, 8471, 8481,
                                                 8491, 8501, 8511])))
        self.assertTrue(np.array_equal(orig_year_set[1986],
                                       np.array([11101, 11111, 11121, 11131, 11141, 11151])))
        self.assertTrue(np.array_equal(orig_year_set[1997],
                                       np.array([12221, 12231, 12241, 12251, 12261, 12271,
                                                 12281, 12291])))
        self.assertTrue(np.array_equal(orig_year_set[2006],
                                       np.array([13571, 13581, 13591, 13601, 13611, 13621,
                                                 13631, 13641, 13651, 13661])))
        self.assertTrue(np.array_equal(orig_year_set[2010],
                                       np.array([14071, 14081, 14091, 14101, 14111, 14121,
                                                 14131, 14141, 14151, 14161, 14171, 14181,
                                                 14191, 14201, 14211, 14221, 14231, 14241,
                                                 14251])))

class TestReaderExcel(unittest.TestCase):
    """Test reader functionality of the Hazard class"""

    def test_hazard_pass(self):
        """Read an hazard excel file correctly."""

        # Read demo excel file
        description = 'One single file.'
        hazard = Hazard.from_excel(HAZ_TEMPLATE_XLS, description=description, haz_type='TC')

        # Check results
        n_events = 100
        n_centroids = 45

        self.assertEqual(hazard.units, '')

        self.assertEqual(hazard.centroids.coord.shape, (n_centroids, 2))
        self.assertEqual(hazard.centroids.coord[0][0], -25.95)
        self.assertEqual(hazard.centroids.coord[0][1], 32.57)
        self.assertEqual(hazard.centroids.coord[n_centroids - 1][0], -24.7)
        self.assertEqual(hazard.centroids.coord[n_centroids - 1][1], 33.88)

        self.assertEqual(len(hazard.event_name), 100)
        self.assertEqual(hazard.event_name[12], 'event013')

        self.assertEqual(hazard.event_id.dtype, int)
        self.assertEqual(hazard.event_id.shape, (n_events,))
        self.assertEqual(hazard.event_id[0], 1)
        self.assertEqual(hazard.event_id[n_events - 1], 100)

        self.assertEqual(hazard.date.dtype, int)
        self.assertEqual(hazard.date.shape, (n_events,))
        self.assertEqual(hazard.date[0], 675874)
        self.assertEqual(hazard.date[n_events - 1], 676329)

        self.assertEqual(hazard.event_name[0], 'event001')
        self.assertEqual(hazard.event_name[50], 'event051')
        self.assertEqual(hazard.event_name[-1], 'event100')

        self.assertEqual(hazard.frequency.dtype, float)
        self.assertEqual(hazard.frequency.shape, (n_events,))
        self.assertEqual(hazard.frequency[0], 0.01)
        self.assertEqual(hazard.frequency[n_events - 2], 0.001)

        self.assertEqual(hazard.intensity.dtype, float)
        self.assertEqual(hazard.intensity.shape, (n_events, n_centroids))

        self.assertEqual(hazard.fraction.dtype, float)
        self.assertEqual(hazard.fraction.shape, (n_events, n_centroids))
        self.assertEqual(hazard.fraction[0, 0], 1)
        self.assertEqual(hazard.fraction[10, 19], 1)
        self.assertEqual(hazard.fraction[n_events - 1, n_centroids - 1], 1)

        self.assertTrue(np.all(hazard.orig))

        # tag hazard
        self.assertEqual(hazard.tag.file_name, HAZ_TEMPLATE_XLS)
        self.assertEqual(hazard.tag.description, description)
        self.assertEqual(hazard.tag.haz_type, 'TC')

class TestReaderMat(unittest.TestCase):
    """Test reader functionality of the ExposuresExcel class"""

    def test_hazard_pass(self):
        """Read a hazard mat file correctly."""
        # Read demo excel file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        # Check results
        n_events = 14450
        n_centroids = 100

        self.assertEqual(hazard.units, 'm/s')

        self.assertEqual(hazard.centroids.coord.shape, (n_centroids, 2))

        self.assertEqual(hazard.event_id.dtype, int)
        self.assertEqual(hazard.event_id.shape, (n_events,))

        self.assertEqual(hazard.frequency.dtype, float)
        self.assertEqual(hazard.frequency.shape, (n_events,))

        self.assertEqual(hazard.intensity.dtype, float)
        self.assertEqual(hazard.intensity.shape, (n_events, n_centroids))
        self.assertEqual(hazard.intensity[12, 46], 12.071393519949979)
        self.assertEqual(hazard.intensity[13676, 49], 17.228323602220616)

        self.assertEqual(hazard.fraction.dtype, float)
        self.assertEqual(hazard.fraction.shape, (n_events, n_centroids))
        self.assertEqual(hazard.fraction[8454, 98], 1)
        self.assertEqual(hazard.fraction[85, 54], 0)

        self.assertEqual(len(hazard.event_name), n_events)
        self.assertEqual(hazard.event_name[124], 125)

        self.assertEqual(len(hazard.date), n_events)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[0]).year, 1851)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[0]).month, 6)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[0]).day, 25)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[78]).year, 1852)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[78]).month, 9)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[78]).day, 22)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[-1]).year, 2011)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[-1]).month, 11)
        self.assertEqual(dt.datetime.fromordinal(hazard.date[-1]).day, 6)

        self.assertTrue(hazard.orig[0])
        self.assertTrue(hazard.orig[11580])
        self.assertTrue(hazard.orig[4940])
        self.assertFalse(hazard.orig[3551])
        self.assertFalse(hazard.orig[10651])
        self.assertFalse(hazard.orig[4818])

        # tag hazard
        self.assertEqual(hazard.tag.file_name, str(HAZ_TEST_MAT))
        self.assertEqual(hazard.tag.description,
                         ' TC hazard event set, generated 14-Nov-2017 10:09:05')
        self.assertEqual(hazard.tag.haz_type, 'TC')

class TestHDF5(unittest.TestCase):
    """Test reader functionality of the ExposuresExcel class"""

    def test_write_read_pass(self):
        """Read a hazard mat file correctly."""
        file_name = str(DATA_DIR.joinpath('test_haz.h5'))

        # Read demo matlab file
        hazard = Hazard.from_mat(HAZ_TEST_MAT)
        hazard.event_name = list(map(str, hazard.event_name))
        for todense_flag in [False, True]:
            if todense_flag:
                hazard.write_hdf5(file_name, todense=todense_flag)
            else:
                hazard.write_hdf5(file_name)

            haz_read = Hazard.from_hdf5(file_name)

            self.assertEqual(str(hazard.tag.file_name), haz_read.tag.file_name)
            self.assertIsInstance(haz_read.tag.file_name, str)
            self.assertEqual(hazard.tag.haz_type, haz_read.tag.haz_type)
            self.assertIsInstance(haz_read.tag.haz_type, str)
            self.assertEqual(hazard.tag.description, haz_read.tag.description)
            self.assertIsInstance(haz_read.tag.description, str)
            self.assertEqual(hazard.units, haz_read.units)
            self.assertIsInstance(haz_read.units, str)
            self.assertTrue(np.array_equal(hazard.centroids.coord, haz_read.centroids.coord))
            self.assertTrue(u_coord.equal_crs(hazard.centroids.crs, haz_read.centroids.crs))
            self.assertTrue(np.array_equal(hazard.event_id, haz_read.event_id))
            self.assertTrue(np.array_equal(hazard.frequency, haz_read.frequency))
            self.assertTrue(np.array_equal(hazard.event_name, haz_read.event_name))
            self.assertIsInstance(haz_read.event_name, list)
            self.assertIsInstance(haz_read.event_name[0], str)
            self.assertTrue(np.array_equal(hazard.date, haz_read.date))
            self.assertTrue(np.array_equal(hazard.orig, haz_read.orig))
            self.assertTrue(np.array_equal(hazard.intensity.toarray(),
                                           haz_read.intensity.toarray()))
            self.assertIsInstance(haz_read.intensity, sparse.csr_matrix)
            self.assertTrue(np.array_equal(hazard.fraction.toarray(), haz_read.fraction.toarray()))
            self.assertIsInstance(haz_read.fraction, sparse.csr_matrix)

class TestCentroids(unittest.TestCase):
    """Test return period statistics"""

    def test_reproject_raster_pass(self):
        """Test reproject_raster reference."""
        haz_fl = Hazard.from_raster([HAZ_DEMO_FL])
        haz_fl.check()

        haz_fl.reproject_raster(dst_crs='epsg:2202')

        self.assertEqual(haz_fl.intensity.shape, (1, 1046408))
        self.assertIsInstance(haz_fl.intensity, sparse.csr_matrix)
        self.assertIsInstance(haz_fl.fraction, sparse.csr_matrix)
        self.assertEqual(haz_fl.fraction.shape, (1, 1046408))
        self.assertTrue(u_coord.equal_crs(haz_fl.centroids.meta['crs'], 'epsg:2202'))
        self.assertEqual(haz_fl.centroids.meta['width'], 968)
        self.assertEqual(haz_fl.centroids.meta['height'], 1081)
        self.assertEqual(haz_fl.fraction.min(), 0)
        self.assertEqual(haz_fl.fraction.max(), 1)
        self.assertEqual(haz_fl.intensity.min(), -9999)
        self.assertTrue(haz_fl.intensity.max() < 4.7)

    def test_raster_to_vector_pass(self):
        """Test raster_to_vector method"""
        haz_fl = Hazard.from_raster([HAZ_DEMO_FL], haz_type='FL')
        haz_fl.check()
        meta_orig = haz_fl.centroids.meta
        inten_orig = haz_fl.intensity
        fract_orig = haz_fl.fraction

        haz_fl.raster_to_vector()

        self.assertEqual(haz_fl.centroids.meta, dict())
        self.assertAlmostEqual(haz_fl.centroids.lat.min(),
                               meta_orig['transform'][5]
                               + meta_orig['height'] * meta_orig['transform'][4]
                               - meta_orig['transform'][4] / 2)
        self.assertAlmostEqual(haz_fl.centroids.lat.max(),
                               meta_orig['transform'][5] + meta_orig['transform'][4] / 2)
        self.assertAlmostEqual(haz_fl.centroids.lon.max(),
                               meta_orig['transform'][2]
                               + meta_orig['width'] * meta_orig['transform'][0]
                               - meta_orig['transform'][0] / 2)
        self.assertAlmostEqual(haz_fl.centroids.lon.min(),
                               meta_orig['transform'][2] + meta_orig['transform'][0] / 2)
        self.assertTrue(u_coord.equal_crs(haz_fl.centroids.crs, meta_orig['crs']))
        self.assertTrue(np.allclose(haz_fl.intensity.data, inten_orig.data))
        self.assertTrue(np.allclose(haz_fl.fraction.data, fract_orig.data))

    def test_reproject_vector_pass(self):
        """Test reproject_vector"""
        haz_fl = Hazard('FL')
        haz_fl.event_id = np.array([1])
        haz_fl.date = np.array([1])
        haz_fl.frequency = np.array([1])
        haz_fl.orig = np.array([1])
        haz_fl.event_name = ['1']
        haz_fl.intensity = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]))
        haz_fl.fraction = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]) / 2)
        haz_fl.centroids = Centroids.from_lat_lon(np.array([1, 2, 3]), np.array([1, 2, 3]))
        haz_fl.check()

        haz_fl.reproject_vector(dst_crs='epsg:2202')
        self.assertTrue(np.allclose(haz_fl.centroids.lat,
                                    np.array([331585.4099637291, 696803.88, 1098649.44])))
        self.assertTrue(np.allclose(haz_fl.centroids.lon,
                                    np.array([11625664.37925186, 11939560.43, 12244857.13])))
        self.assertTrue(u_coord.equal_crs(haz_fl.centroids.crs, 'epsg:2202'))
        self.assertTrue(np.allclose(haz_fl.intensity.toarray(), np.array([0.5, 0.2, 0.1])))
        self.assertTrue(np.allclose(haz_fl.fraction.toarray(), np.array([0.5, 0.2, 0.1]) / 2))

    def test_vector_to_raster_pass(self):
        """Test vector_to_raster"""
        haz_fl = Hazard('FL')
        haz_fl.event_id = np.array([1])
        haz_fl.date = np.array([1])
        haz_fl.frequency = np.array([1])
        haz_fl.orig = np.array([1])
        haz_fl.event_name = ['1']
        haz_fl.intensity = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]))
        haz_fl.fraction = sparse.csr_matrix(np.array([0.5, 0.2, 0.1]) / 2)
        haz_fl.centroids = Centroids.from_lat_lon(np.array([1, 2, 3]), np.array([1, 2, 3]))
        haz_fl.check()

        haz_fl.vector_to_raster()
        self.assertTrue(u_coord.equal_crs(haz_fl.centroids.meta['crs'], 'epsg:4326'))
        self.assertAlmostEqual(haz_fl.centroids.meta['transform'][0], 1.0)
        self.assertAlmostEqual(haz_fl.centroids.meta['transform'][1], 0)
        self.assertAlmostEqual(haz_fl.centroids.meta['transform'][2], 0.5)
        self.assertAlmostEqual(haz_fl.centroids.meta['transform'][3], 0)
        self.assertAlmostEqual(haz_fl.centroids.meta['transform'][4], -1.0)
        self.assertAlmostEqual(haz_fl.centroids.meta['transform'][5], 3.5)
        self.assertEqual(haz_fl.centroids.meta['height'], 3)
        self.assertEqual(haz_fl.centroids.meta['width'], 3)
        self.assertEqual(haz_fl.centroids.lat.size, 0)
        self.assertEqual(haz_fl.centroids.lon.size, 0)
        self.assertTrue(haz_fl.intensity.min() >= 0)
        self.assertTrue(haz_fl.intensity.max() <= 0.5)
        self.assertTrue(haz_fl.fraction.min() >= 0)
        self.assertTrue(haz_fl.fraction.max() <= 0.5 / 2)


class TestClear(unittest.TestCase):
    """Test clear method"""

    def test_clear(self):
        """Clear method clears everything"""
        haz1 = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        haz1.units = "m"
        haz1.foo = np.arange(10)
        haz1.clear()
        self.assertEqual(list(vars(haz1.tag).values()), ['', '', ''])
        self.assertEqual(haz1.units, '')
        self.assertEqual(haz1.centroids.size, 0)
        self.assertEqual(len(haz1.event_name), 0)
        for attr in vars(haz1).keys():
            if attr not in ['tag', 'units', 'event_name', 'pool']:
                self.assertEqual(getattr(haz1, attr).size, 0)
        self.assertIsNone(haz1.pool)

    def test_clear_pool(self):
        """Clear method should not clear a process pool"""
        haz1 = Hazard.from_excel(HAZ_TEMPLATE_XLS, haz_type='TC')
        pool = Pool(nodes=2)
        haz1.pool = pool
        haz1.check()
        haz1.clear()
        self.assertEqual(haz1.pool, pool)
        pool.close()
        pool.join()
        pool.clear()


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestLoader)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestHDF5))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestReaderExcel))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestReaderMat))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestRemoveDupl))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestSelect))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestStats))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestYearset))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestAppend))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestCentroids))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestClear))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Tag class.
"""

__all__ = ['Tag']

import logging
from pathlib import Path

LOGGER = logging.getLogger(__name__)

class Tag(object):
    """Contain information used to tag a Hazard.

    Attributes
    ----------
    file_name : str or list(str)
        name of the source file(s)
    haz_type : str
        acronym defining the hazard type (e.g. 'TC')
    description : str or list(str)
        description(s) of the data
    """

    def __init__(self, haz_type='', file_name='', description=''):
        """Initialize values.

        Parameters
        ----------
        haz_type : str, optional
            acronym of the hazard type (e.g. 'TC').
        file_name : str or list(str), optional
            file name(s) to read
        description : str or list(str), optional
            description of the data
        """
        self.haz_type = haz_type
        self.file_name = file_name
        self.description = description

    def append(self, tag):
        """Append input Tag instance information to current Tag."""
        if self.haz_type == '':
            self.haz_type = tag.haz_type
        if tag.haz_type != self.haz_type:
            raise ValueError("Hazards of different type can't be appended: %s != %s."
                             % (self.haz_type, tag.haz_type))

        # add file name if not present in tag
        if self.file_name == '':
            self.file_name = tag.file_name
            self.description = tag.description
        elif tag.file_name == '':
            return
        else:
            if not isinstance(self.file_name, list):
                self.file_name = [self.file_name]
            if not isinstance(tag.file_name, list):
                to_add = [tag.file_name]
            else:
                to_add = tag.file_name
            self.file_name.extend(to_add)

            if not isinstance(self.description, list):
                self.description = [self.description]
            if not isinstance(tag.description, list):
                to_add = [tag.description]
            else:
                to_add = tag.description
            self.description.extend(to_add)

    def join_file_names(self):
        """Get a string with the joined file names."""
        if not isinstance(self.file_name, list):
            join_file = Path(self.file_name).stem
        else:
            join_file = ' + '.join([
                Path(single_name).stem
                for single_name in self.file_name
            ])
        return join_file

    def join_descriptions(self):
        """Get a string with the joined descriptions."""
        if not isinstance(self.file_name, list):
            join_desc = self.description
        else:
            join_desc = ' + '.join([file for file in self.description])
        return join_desc

    def __str__(self):
        return ' Type: ' + self.haz_type + '\n File: ' + \
            self.join_file_names() + '\n Description: ' + \
            self.join_descriptions()

    __repr__ = __str__
